from __future__ import annotations

import logging
import sys
from typing import TYPE_CHECKING

import boto3
import click
from botocore.config import Config

from . import model_list
from .util import formatted_print

if TYPE_CHECKING:
    from io import TextIOWrapper

    from bedrock_bot.models.base_model import _BedrockModel

CONTEXT_SETTINGS = {"help_option_names": ["--help", "-h"]}
LOG_FORMATTER = logging.Formatter(
    "%(asctime)s - %(levelname)s - %(module)s - %(message)s",
    "%Y-%m-%d %H:%M:%S",
)
log = None


def configure_logger(*, verbose: bool) -> None:
    global log  # noqa: PLW0603
    log_level = logging.ERROR
    if verbose:
        log_level = logging.INFO

    logging.basicConfig(level=log_level)
    log = logging.getLogger()
    log.handlers[0].setFormatter(LOG_FORMATTER)
    log.info(f"Log level set to {logging.getLevelName(log_level)}")


def available_models() -> list[str]:
    return [x.name for x in model_list]


def model_class_from_input(value: str) -> type[_BedrockModel]:
    try:
        return next(x for x in model_list if x.name.lower() == value.lower())
    except StopIteration as err:
        msg = f"Invalid value: {value}. Allowed values are: {available_models}"
        raise click.BadParameter(msg) from err


def generate_boto_config(region: str) -> Config:
    boto_config = Config()
    if region:
        boto_config = Config(region_name=region)
    elif boto3.setup_default_session() and not boto3.DEFAULT_SESSION.region_name:
        boto_config = Config(region_name="us-east-1")
    return boto_config


def get_user_input(instance: _BedrockModel, args: list[str]) -> str:
    if instance.messages == [] and not sys.stdin.isatty():
        user_input = sys.stdin.read()
        print(f"> {user_input}")  # noqa: T201
    elif instance.messages == [] and args:
        user_input = " ".join(args)
        print(f"> {user_input}")  # noqa: T201
    elif not sys.stdin.isatty():
        print(  # noqa: T201
            "Note that you can only do one-shot requests when providing input via stdin",
        )
        sys.exit()
    else:
        user_input = input("> ")

    return user_input


def handle_input_files(input_file: list[TextIOWrapper]) -> list:
    output = []
    if input_file:
        output = [f"File '{file.name}':\n{file.read()}" for file in input_file]
    return output


@click.command()
@click.argument("args", nargs=-1)
@click.option(
    "-r",
    "--region",
    help="The AWS region to use for requests. If no default region is specified, defaults to us-east-1",
)
@click.option("--raw-output", help="Don't interpret markdown in the AI response")
@click.option(
    "-m",
    "--model",
    type=click.Choice(available_models(), case_sensitive=False),
    default="Claude-3-Haiku",
    help="The model to use for requests",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Enable verbose logging messages",
)
@click.option(
    "-i",
    "--input-file",
    multiple=True,
    type=click.File(),
    help="Read in file(s) to be used in your queries",
)
def main(  # noqa: PLR0913
    *,
    model: str,
    region: str,
    raw_output: str,
    args: list[str],
    verbose: bool,
    input_file: list[TextIOWrapper],
) -> None:
    configure_logger(verbose=verbose)

    model_class = model_class_from_input(model)
    boto_config = generate_boto_config(region)
    instance = model_class(boto_config=boto_config)

    print(  # noqa: T201
        f"Hello! I am an AI assistant powered by Amazon Bedrock and using the model {instance.name}. Enter 'quit' or"
        " 'exit' at any time to exit. How may I help you today?",
    )
    print(  # noqa: T201
        "(You can clear existing context by starting a query with 'new>' or 'reset>')",
    )

    while True:
        print()  # noqa: T201
        user_input = get_user_input(instance, args)

        if not user_input:
            continue
        if user_input.lower() == "quit" or user_input.lower() == "exit":
            print("\nGoodbye!")  # noqa: T201
            sys.exit()
        if user_input.lower().startswith("new>") or user_input.lower().startswith(
            "reset>",
        ):
            print("\nResetting...")  # noqa: T201
            instance.reset()
            continue

        if not instance.messages:
            user_input += "\n"
            user_input += "\n".join(handle_input_files(input_file))

        response = instance.invoke(user_input)

        if raw_output:
            print(response)  # noqa: T201
        else:
            formatted_print(response)

        print()  # noqa: T201



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from rich.console import Console

from .base_model import _BedrockModel

if TYPE_CHECKING:
    from botocore.config import Config

logger = logging.getLogger()

console = Console()


class MistralLarge(_BedrockModel):
    name = "Mistral-Large"
    model_params = {  # noqa: RUF012
        "max_tokens": 200,
        "temperature": 0.5,
        "top_p": 0.9,
        "top_k": 50,
    }

    def __init__(self, boto_config: None | Config = None) -> None:
        self._model_id = "mistral.mistral-large-2402-v1:0"
        super().__init__(
            boto_config=boto_config,
        )

    def _format_messages(self) -> str:
        formatted_messages = ["<s>[INST]"]
        for message in self.messages:
            role = message["role"]
            content = message["content"]
            formatted_message = f"{role}: {content}"
            formatted_messages.append(formatted_message)
        formatted_messages.append("[/INST]")
        return "\n".join(formatted_messages)

    def _create_invoke_body(self) -> dict:
        prompt = self._format_messages()
        return {"prompt": prompt}

    def _handle_response(self, body: dict) -> str:
        response_message = body["outputs"][0]

        return response_message["text"]



from __future__ import annotations

import json
import logging
from enum import Enum

import boto3
from botocore.config import Config
from rich.console import Console

logger = logging.getLogger()

console = Console()


class ConversationRole(Enum):
    USER = "user"
    ASSISTANT = "assistant"

    def __str__(self) -> str:
        return self.value


class _BedrockModel:
    """Base class for all models.

    Usage:
    Add a message to the conversation and invoke the model with `model.invoke(message)`.

    Creating new implementations of this base model:
    - Define extra params to inject to the invoke call on self.model_params
    - Provide a model_id to `super().__init__()`
    - If your model needs a different body than the default, overwrite `self._create_invoke_body()`
    """

    name = "Base Model"
    model_params = {}  # noqa: RUF012
    _model_id: str

    def __init__(
        self,
        boto_config: None | Config = None,
    ) -> None:
        if not boto_config:
            boto_config = Config()
        self._bedrock = boto3.client("bedrock-runtime", config=boto_config)
        self.messages = []

    def reset(self) -> None:
        self.messages = []

    def append_message(self, role: ConversationRole, message: str) -> None:
        self.messages.append({"role": role.value, "content": message})

    def invoke(self, message: str) -> str:
        self.append_message(ConversationRole.USER, message)

        response = self._invoke()
        self.append_message(ConversationRole.ASSISTANT, response)
        return response

    def _create_invoke_body(self) -> dict:
        raise NotImplementedError

    def _handle_response(self, body: dict) -> str:
        raise NotImplementedError

    def _invoke(self) -> str:
        body = self._create_invoke_body() | self.model_params

        with console.status("[bold green]Waiting for response..."):
            logger.info(f"Sending current messages to AI: {self.messages}")
            response = self._bedrock.invoke_model(
                modelId=self._model_id,
                body=json.dumps(body),
            )
            body = json.loads(response["body"].read())

            return self._handle_response(body)



from __future__ import annotations

import logging
from datetime import datetime
from typing import TYPE_CHECKING

from rich.console import Console

from .base_model import _BedrockModel

if TYPE_CHECKING:
    from botocore.config import Config

logger = logging.getLogger()

console = Console()


class _Claude3(_BedrockModel):
    date_stamp = datetime.now().strftime("%A, %B %d, %Y")  # noqa: DTZ005
    system_prompt = f"""The assistant is Claude, created by Anthropic.
The current date is {date_stamp}.

Claude's knowledge base was last updated on August 2023. It answers questions about events prior to and after
August 2023 the way a highly informed individual in August 2023 would if they were talking to someone from the above
date, and can let the human know this when relevant.

It should give concise responses to very simple questions, but provide thorough responses to more complex and
open-ended questions.

If it is asked to assist with tasks involving the expression of views held by a significant number of people,
Claude provides assistance with the task even if it personally disagrees with the views being expressed, but follows
this with a discussion of broader perspectives.

Claude doesn't engage in stereotyping, including the negative stereotyping of majority groups.

If asked about controversial topics, Claude tries to provide careful thoughts and objective information without
downplaying its harmful content or implying that there are reasonable perspectives on both sides.

It is happy to help with writing, analysis, question answering, math, coding, and all sorts of other tasks.
It uses markdown for coding.

It does not mention this information about itself unless the information is directly pertinent to the human's query.
""".replace("\n", " ")
    model_params = {  # noqa: RUF012
        "max_tokens": 2000,
        "temperature": 1,
        "top_k": 250,
        "top_p": 0.999,
        "stop_sequences": ["\n\nHuman:"],
        "anthropic_version": "bedrock-2023-05-31",
        "system": system_prompt,
    }

    def _create_invoke_body(self) -> dict:
        return {
            "messages": self.messages,
        }

    def _handle_response(self, body: dict) -> str:
        response_message = body["content"][0]

        if response_message["type"] != "text":
            raise RuntimeError("Unexpected response type to prompt: " + response_message["type"])

        return response_message["text"]

    def __init__(self, boto_config: None | Config = None) -> None:
        super().__init__(
            boto_config=boto_config,
        )


class Claude3Sonnet(_Claude3):
    name = "Claude-3-Sonnet"

    def __init__(self, boto_config: None | Config = None) -> None:
        self._model_id = "anthropic.claude-3-sonnet-20240229-v1:0"

        super().__init__(
            boto_config=boto_config,
        )


class Claude3Haiku(_Claude3):
    name = "Claude-3-Haiku"

    def __init__(self, boto_config: None | Config = None) -> None:
        self._model_id = "anthropic.claude-3-haiku-20240307-v1:0"

        super().__init__(
            boto_config=boto_config,
        )
